import { AbortedError, getGraphStatistics, GPError, graphQueryBackend2graphQuery, type DbConnection, type QueryRequest } from 'ts-common';

import { QUERY_CACHE_DURATION, rabbitMq, redis, ums, type QueryExecutionTypes } from '../variables';
import { log } from '../logger';
import { QueryPublisher } from '../utils/queryPublisher';
import { query2Cypher } from '../utils/cypher/converter';
import { parseCountCypherQuery, parseCypherQuery } from '../utils/cypher/queryParser';
import { formatTimeDifference } from 'ts-common/src/logger/logger';
import { Query2BackendQuery } from '../utils/reactflow/query2backend';
import type { GraphQueryCountResultFromBackend, GraphQueryResultMetaFromBackend } from 'ts-common/src/model/webSocket/graphResult';
import { RabbitMqBroker } from 'ts-common/rabbitMq';
import { Neo4jConnection } from 'ts-common/neo4j';
import { Neo4jError } from 'neo4j-driver';

async function cacheCheck<T>(cacheKey: string): Promise<T | undefined> {
  log.debug('Checking cache for query, with cache ttl', QUERY_CACHE_DURATION, 'seconds');
  const cached = await redis.client.get(cacheKey);
  if (cached) {
    log.info('Cache hit for query');
    const buf = Buffer.from(cached, 'base64');
    const inflated = Bun.gunzipSync(new Uint8Array(buf));
    const dec = new TextDecoder();
    const cachedMessage = JSON.parse(dec.decode(inflated)) as T;
    return cachedMessage;
  }
}

export const queryService = async (
  db: DbConnection,
  cypher: string,
  useCached: boolean,
  sessionID: string | undefined = undefined,
  callID: string | undefined = undefined,
): Promise<GraphQueryResultMetaFromBackend> => {
  let index = 0;
  const disambiguatedQuery = cypher.replace(/\d{13}/g, () => (index++).toString());
  const cacheKey = Bun.hash(JSON.stringify({ db: db, query: disambiguatedQuery })).toString();

  if (QUERY_CACHE_DURATION === '') {
    log.info('Query cache disabled, skipping cache check');
  } else if (!useCached) {
    log.info('Skipping cache check for query due to parameter', useCached);
  } else {
    const cachedMessage = await cacheCheck<GraphQueryResultMetaFromBackend>(cacheKey);
    if (cachedMessage) {
      log.debug('Cache hit for query', disambiguatedQuery);
      return cachedMessage;
    }
  }

  // TODO: only neo4j is supported for now
  const connection = new Neo4jConnection(db);
  try {
    const [neo4jResult] =
      sessionID != null && callID != null
        ? await connection.runWithAbort(redis, [cypher], sessionID, callID, 5000, false)
        : await connection.run([cypher]);
    const graph = graphQueryBackend2graphQuery(parseCypherQuery(neo4jResult.records));

    // calculate metadata
    const result: GraphQueryResultMetaFromBackend = { ...graph, metaData: getGraphStatistics(graph) };

    // Force garbage collection
    neo4jResult.records = [];
    Bun.gc(true);

    // cache result
    if (QUERY_CACHE_DURATION !== '') {
      log.info('Started gzipping...');
      const compressedMessage = Bun.gzipSync(JSON.stringify(result));
      log.info('Done gzipping, started encoding to base64...');
      const base64Message = Buffer.from(compressedMessage).toString('base64');
      log.info('Done encoding, sending to redis...');

      // if cache enabled, cache the result
      await redis.setWithExpire(cacheKey, base64Message, QUERY_CACHE_DURATION); // ttl in seconds
      log.info('cached in redis');
    }

    return result;
  } catch (error) {
    if (error instanceof AbortedError) {
      log.info('Query aborted:', cypher);
      throw error;
    } else if (error instanceof Neo4jError) {
      log.error('Error in queryServiceReader', error.message, error);
      throw new GPError('Error querying neo4j', error);
    } else {
      log.error('Error parsing query result:', cypher, error);
      throw new GPError('Error parsing query result', error);
    }
  } finally {
    connection.close();
  }
};

export const queryCountService = async (
  db: DbConnection,
  cypher: string,
  useCached: boolean,
  sessionID: string,
  callID: string,
): Promise<GraphQueryCountResultFromBackend> => {
  let index = 0;
  const disambiguatedQuery = cypher.replace(/\d{13}/g, () => (index++).toString());
  const cacheKey = Bun.hash(JSON.stringify({ db: db, query: disambiguatedQuery })).toString();

  if (QUERY_CACHE_DURATION === '') {
    log.info('Query cache disabled, skipping cache check');
  } else if (!useCached) {
    log.info('Skipping cache check for query due to parameter', useCached);
  } else {
    const cachedMessage = await cacheCheck<GraphQueryCountResultFromBackend>(cacheKey);
    if (cachedMessage) {
      log.debug('Cache hit for query', disambiguatedQuery);
      return cachedMessage;
    }
  }

  // TODO: only neo4j is supported for now
  const connection = new Neo4jConnection(db);
  try {
    const [neo4jCountResult] = await connection.runWithAbort(redis, [cypher], sessionID, callID, 5000, false);
    const countGraph = { nodeCounts: parseCountCypherQuery(neo4jCountResult.records) };

    // cache result
    if (QUERY_CACHE_DURATION !== '') {
      log.info('Started gzipping...');
      const compressedMessage = Bun.gzipSync(JSON.stringify(countGraph));
      log.info('Done gzipping, started encoding to base64...');
      const base64Message = Buffer.from(compressedMessage).toString('base64');
      log.info('Done encoding, sending to redis...');

      // if cache enabled, cache the result
      await redis.setWithExpire(cacheKey, base64Message, QUERY_CACHE_DURATION); // ttl in seconds
      log.info('cached in redis');
    }

    return countGraph;
  } catch (error) {
    if (error instanceof AbortedError) {
      log.info('Count Query aborted:', cypher);
      throw error;
    } else if (error instanceof Neo4jError) {
      log.error('Error in queryServiceReader', error.message, error);
      throw new GPError('Error querying neo4j', error);
    } else {
      log.error('Error parsing count query result:', cypher, error);
      throw new GPError('Error parsing count query result', error);
    }
  } finally {
    connection.close();
  }
};

export const queryServiceReader = async (frontendPublisher: RabbitMqBroker, mlPublisher: RabbitMqBroker, type: QueryExecutionTypes) => {
  if (type == null) {
    log.error('Unsupported query execution type:', type);
    throw new Error('Unsupported query execution type');
  }
  log.info('Starting query reader for', type);

  const queryServiceConsumer = await new RabbitMqBroker(
    rabbitMq,
    'requests-exchange',
    `${type}-query-queue`,
    `${type}-query-request`,
  ).connect();

  log.info('Connected to RabbitMQ ST!');

  await queryServiceConsumer.startConsuming<QueryRequest>('query-service', async (message, headers) => {
    if (message.queryID == null) {
      log.error('QueryID not set in message:', message.queryID);
      return;
    }

    const publisher = new QueryPublisher(frontendPublisher, mlPublisher, headers, message.queryID);

    try {
      const ss = await ums.getUserSaveState(headers.message.sessionData.userID, message.saveStateID);

      if (!ss) {
        log.error('Invalid SaveState received in queryServiceConsumer:', ss);
        throw new GPError('Invalid SaveState');
      }

      log.debug('Received query request:', message, headers, ss);
      log.debug('Received routing key:', headers.routingKey);

      publisher.publishStatusToFrontend('Received');

      if (ss == null || ss.dbConnections == null || ss.dbConnections[0] == null || ss.dbConnections.length === 0) {
        log.error('Invalid SaveState received in queryServiceConsumer:', ss);
        throw new GPError('Invalid SaveState');
      }

      let activeQuery = ss.queryStates.activeQueryId;
      if (message.queryID) {
        if (ss.queryStates.openQueryArray.find(q => q.id === message.queryID) == null) {
          log.error('Query not found in SaveState:', message.queryID, ss.queryStates.openQueryArray);
          throw new GPError('Query not found');
        }
        activeQuery = message.queryID;
      }

      if (activeQuery == null || activeQuery == -1) {
        log.error('No active query found in SaveState:', ss);
        throw new GPError('No active query found');
      }

      const activeQueryInfo = ss.queryStates.openQueryArray.find(q => q.id === activeQuery);
      if (activeQueryInfo == null) {
        log.error('Active query not found in SaveState:', ss.queryStates.activeQueryId, ss.queryStates.openQueryArray);
        throw new GPError('Active query not found');
      }

      const visualQuery = activeQueryInfo.graph; //ss.queries[0].graph;
      log.debug('Received query request:', message, headers, visualQuery);
      if (visualQuery.nodes.length === 0) {
        log.info('Empty query received');
        publisher.publishResultToFrontend({
          nodes: [],
          edges: [],
          metaData: {
            topological: { density: 0, self_loops: 0 },
            nodes: { count: 0, labels: [], types: {} },
            edges: { count: 0, labels: [], types: {} },
          },
        });
        publisher.publishCountResultToFrontend({ nodeCounts: { updatedAt: 0 } });
        return;
      }

      const queryBuilderSettings = activeQueryInfo.settings; //ss.queries[0].settings;
      const ml = message.ml;
      let startTime = Date.now();
      const convertedQuery = Query2BackendQuery(ss.id, visualQuery, queryBuilderSettings, ml);
      log.info(`Query converted in ${formatTimeDifference(Date.now() - startTime)}`);

      log.debug('translating query:', convertedQuery);
      publisher.publishStatusToFrontend('Translating');

      const cypher = query2Cypher(convertedQuery);
      const query = cypher.query;
      if (query == null) {
        throw new GPError(`Error translating query ${convertedQuery}`);
      }

      log.debug('Translated query FROM:', convertedQuery);
      log.info('Translated query:', query);
      log.info('Translated query:', cypher.countQuery);
      publisher.publishTranslationResultToFrontend(query);

      for (let i = 0; i < ss.dbConnections.length; i++) {
        let result: GraphQueryResultMetaFromBackend;
        try {
          log.info('Executing query on database');
          startTime = Date.now();
          result = await queryService(
            ss.dbConnections[i],
            cypher.query,
            message.useCached,
            headers.message.sessionData.sessionID,
            headers.callID,
          );
          publisher.publishResultToFrontend(result);
          log.info(`Query executed in ${formatTimeDifference(Date.now() - startTime)}`);
        } catch (error) {
          if (error instanceof AbortedError) {
            publisher.publishResultToFrontend({} as any, 'aborted');
            return;
          } else {
            throw new GPError('Error querying database', error);
          }
        }

        let countResult: GraphQueryCountResultFromBackend;
        try {
          startTime = Date.now();
          countResult = await queryCountService(
            ss.dbConnections[i],
            cypher.countQuery,
            message.useCached,
            headers.message.sessionData.sessionID,
            headers.callID,
          );
          publisher.publishCountResultToFrontend(countResult);
          log.info(`Query Count executed in ${formatTimeDifference(Date.now() - startTime)}`);
        } catch (error) {
          if (error instanceof AbortedError) {
            publisher.publishCountResultToFrontend({ nodeCounts: { updatedAt: 0 } }, 'aborted');
            return;
          } else {
            throw new GPError('Error querying database', error);
          }
        }

        try {
          // Cache nodeCounts such that we can display differentiation for each query
          await ums.updateQuery(ss.userId, message.saveStateID, {
            ...activeQueryInfo,
            graph: {
              ...activeQueryInfo.graph,
            },
            graphCounts: countResult,
          });

          if (convertedQuery.machineLearning && convertedQuery.machineLearning.length > 0) {
            for (let i = 0; i < convertedQuery.machineLearning.length; i++) {
              try {
                publisher.publishMachineLearningRequest(result, convertedQuery.machineLearning[i], headers);
                log.debug('Published machine learning request', convertedQuery.machineLearning[i]);
              } catch (error) {
                throw new GPError('Error publishing machine learning request', error);
              }
            }
          }
        } catch (error) {
          throw new GPError('Error during database query execution', error);
        }

        Bun.gc(true);
      }
    } catch (e: any) {
      // Clean resolve
      if (e instanceof GPError) {
        log.error('Error in queryServiceReader', e.message, e);
        e.log();
        publisher.publishErrorToFrontend(e);
      }
      if (e instanceof Neo4jError) {
        log.error('Error in queryServiceReader', e.message, e);
        publisher.publishErrorToFrontend(new GPError(e.message, e));
      } else {
        log.error('Error in queryServiceReader on an uncaught error!', e.message, e);
        publisher.publishErrorToFrontend(new GPError(e.message, e));
      }
    } finally {
      setTimeout(() => Bun.gc(true), 100);
    }
  });
};
