import { GPError, wsReturnKey, type BackendMessageHeader, type MachineLearning, type ToMLMessage } from 'ts-common';
import { log } from '../logger';
import type {
  GraphQueryCountResultFromBackend,
  GraphQueryResultFromBackend,
  GraphQueryResultMetaFromBackend,
} from 'ts-common/src/model/webSocket/graphResult';
import type { RabbitMqBroker } from 'ts-common/rabbitMq';

export class QueryPublisher {
  private frontendPublisher: RabbitMqBroker;
  private mlPublisher: RabbitMqBroker;
  private routingKey: string;
  private headers: BackendMessageHeader;
  private queryID: number;

  constructor(frontendPublisher: RabbitMqBroker, mlPublisher: RabbitMqBroker, headers: BackendMessageHeader, queryID: number) {
    this.frontendPublisher = frontendPublisher;
    this.mlPublisher = mlPublisher;
    this.headers = headers;
    this.routingKey = headers.routingKey;
    this.queryID = queryID;
  }

  publishStatusToFrontend(status: string) {
    this.frontendPublisher.publishMessageToFrontend(
      {
        type: wsReturnKey.queryStatusUpdate,
        callID: this.headers.callID,
        value: this.queryID,
        status: status,
      },
      this.routingKey,
      this.headers,
    );
  }

  publishErrorToFrontend(error: GPError) {
    this.frontendPublisher.publishMessageToFrontend(
      {
        type: wsReturnKey.queryStatusError,
        callID: this.headers.callID,
        value: { ...error.toJSON(), extra: { queryID: this.queryID } },
        status: error.message,
      },
      this.routingKey,
      this.headers,
    );
  }

  publishTranslationResultToFrontend(query: string) {
    this.frontendPublisher.publishMessageToFrontend(
      {
        type: wsReturnKey.queryStatusTranslationResult,
        callID: this.headers.callID,
        value: {
          result: query,
          queryID: this.queryID,
        },
        status: 'success',
      },
      this.routingKey,
      this.headers,
    );
  }

  publishResultToFrontend(result: GraphQueryResultMetaFromBackend, status: string = 'success') {
    this.frontendPublisher.publishMessageToFrontend(
      {
        type: wsReturnKey.queryStatusResult,
        callID: this.headers.callID,
        value: {
          result: {
            type: 'nodelink',
            payload: result,
          },
          queryID: this.queryID,
        },
        status: status,
      },
      this.routingKey,
      this.headers,
    );
  }

  publishCountResultToFrontend(result: GraphQueryCountResultFromBackend, status: string = 'success') {
    this.frontendPublisher.publishMessageToFrontend(
      {
        type: wsReturnKey.queryCountResult,
        callID: this.headers.callID,
        value: {
          result: {
            type: 'count',
            payload: result,
          },
          queryID: this.queryID,
        },
        status: status,
      },
      this.routingKey,
      this.headers,
    );
  }

  publishMachineLearningRequest(result: GraphQueryResultFromBackend, mlAttributes: MachineLearning, headers: BackendMessageHeader) {
    // FIXME: Change ML to use the same message format that the frontend uses
    const toMlResult = {
      nodes: result.nodes.map(node => ({ ...node, id: node._id })),
      edges: result.edges.map(edge => ({ ...edge, id: edge._id })),
    };

    const message: ToMLMessage = {
      queryID: headers.callID,
      type: 'query_result',
      value: {
        type: 'nodelink',
        payload: toMlResult,
      },
      mlAttributes: mlAttributes.parameters,
    };

    const queueName = mlType2Queue[mlAttributes.type];
    if (!queueName) {
      log.error('Invalid ML type:', mlAttributes, mlAttributes.type);
      throw new Error('Invalid ML type');
    }

    this.mlPublisher.publishMessage({
      message: JSON.stringify(message),
      queueName,
      routingKey: queueName,
      options: {
        headers: {
          ...headers,
          message: Buffer.from(JSON.stringify(headers.message)),
        },
      },
    });

    log.debug('Published ML request:', queueName, headers.callID);
  }
}

const mlType2Queue = {
  centrality: 'ctr_queue',
  linkPrediction: 'lpr_queue',
  communityDetection: 'cdt_queue',
  shortestPath: 'stp_queue',
};
